@@ -1,6 +1,6 @@
 //========================================================================
 // Context creation and information tool
-// Copyright (c) Camilla Berglund <elmindreda@elmindreda.org>
+// Copyright (c) Camilla Berglund <elmindreda@glfw.org>
 //
 // This software is provided 'as-is', without any express or implied
 // warranty. In no event will the authors be held liable for any damages
@@ -23,14 +23,11 @@
 //
 //========================================================================
 
+#define VK_NO_PROTOTYPES
+#include <vulkan/vulkan.h>
+#include <glad/glad.h>
 #include <GLFW/glfw3.h>
 
-/* HACK: Explicitly include the glext.h shipping with GLFW, as this program uses
- *       many modern symbols not provided by the versions in some development
- *       environments (for example on OS X).
- */
-#include <GL/glext.h>
-
 #include <stdio.h>
 #include <stdlib.h>
 #include <string.h>
@@ -44,6 +41,9 @@
 #define API_NAME_OPENGL     "gl"
 #define API_NAME_OPENGL_ES  "es"
 
+#define API_NAME_NATIVE     "native"
+#define API_NAME_EGL        "egl"
+
 #define PROFILE_NAME_CORE   "core"
 #define PROFILE_NAME_COMPAT "compat"
 
@@ -63,10 +63,14 @@ static void usage(void)
     printf("  -b, --behavior=BEHAVIOR   the release behavior to use ("
                                         BEHAVIOR_NAME_NONE " or "
                                         BEHAVIOR_NAME_FLUSH ")\n");
+    printf("  -c, --context-api=API     the context creation API to use ("
+                                        API_NAME_NATIVE " or "
+                                        API_NAME_EGL ")\n");
     printf("  -d, --debug               request a debug context\n");
     printf("  -f, --forward             require a forward-compatible context\n");
     printf("  -h, --help                show this help\n");
-    printf("  -l, --list-extensions     list all client API extensions\n");
+    printf("  -l, --list-extensions     list all Vulkan and client API extensions\n");
+    printf("      --list-layers         list all Vulkan layers\n");
     printf("  -m, --major=MAJOR         the major number of the required "
                                         "client API version\n");
     printf("  -n, --minor=MINOR         the minor number of the required "
@@ -93,6 +97,7 @@ static void usage(void)
     printf("      --stereo              request stereo rendering\n");
     printf("      --srgb                request an sRGB capable framebuffer\n");
     printf("      --singlebuffer        request single-buffering\n");
+    printf("      --no-error            request a context that does not emit errors\n");
 }
 
 static void error_callback(int error, const char* description)
@@ -100,6 +105,22 @@ static void error_callback(int error, const char* description)
     fprintf(stderr, "Error: %s\n", description);
 }
 
+static const char* get_device_type_name(VkPhysicalDeviceType type)
+{
+    if (type == VK_PHYSICAL_DEVICE_TYPE_OTHER)
+        return "other";
+    else if (type == VK_PHYSICAL_DEVICE_TYPE_INTEGRATED_GPU)
+        return "integrated GPU";
+    else if (type == VK_PHYSICAL_DEVICE_TYPE_DISCRETE_GPU)
+        return "discrete GPU";
+    else if (type == VK_PHYSICAL_DEVICE_TYPE_VIRTUAL_GPU)
+        return "virtual GPU";
+    else if (type == VK_PHYSICAL_DEVICE_TYPE_CPU)
+        return "CPU";
+
+    return "unknown";
+}
+
 static const char* get_api_name(int api)
 {
     if (api == GLFW_OPENGL_API)
@@ -150,47 +171,161 @@ static const char* get_strategy_name_glfw(int strategy)
     return "unknown";
 }
 
-static void list_extensions(int api, int major, int minor)
+static void list_context_extensions(int client, int major, int minor)
 {
     int i;
     GLint count;
     const GLubyte* extensions;
 
-    printf("%s context supported extensions:\n", get_api_name(api));
+    printf("%s context extensions:\n", get_api_name(client));
 
-    if (api == GLFW_OPENGL_API && major > 2)
+    if (client == GLFW_OPENGL_API && major > 2)
     {
-        PFNGLGETSTRINGIPROC glGetStringi =
-            (PFNGLGETSTRINGIPROC) glfwGetProcAddress("glGetStringi");
-        if (!glGetStringi)
-        {
-            glfwTerminate();
-            exit(EXIT_FAILURE);
-        }
-
         glGetIntegerv(GL_NUM_EXTENSIONS, &count);
 
         for (i = 0;  i < count;  i++)
-            puts((const char*) glGetStringi(GL_EXTENSIONS, i));
+            printf(" %s\n", (const char*) glGetStringi(GL_EXTENSIONS, i));
     }
     else
     {
         extensions = glGetString(GL_EXTENSIONS);
         while (*extensions != '\0')
         {
-            if (*extensions == ' ')
-                putchar('\n');
-            else
-                putchar(*extensions);
+            putchar(' ');
 
-            extensions++;
+            while (*extensions != '\0' && *extensions != ' ')
+            {
+                putchar(*extensions);
+                extensions++;
+            }
+
+            while (*extensions == ' ')
+                extensions++;
+
+            putchar('\n');
         }
     }
-
-    putchar('\n');
 }
 
-static GLboolean valid_version(void)
+static void list_vulkan_instance_extensions(void)
+{
+    uint32_t i, ep_count = 0;
+    VkExtensionProperties* ep;
+    PFN_vkEnumerateInstanceExtensionProperties vkEnumerateInstanceExtensionProperties =
+        (PFN_vkEnumerateInstanceExtensionProperties)
+        glfwGetInstanceProcAddress(NULL, "vkEnumerateInstanceExtensionProperties");
+
+    printf("Vulkan instance extensions:\n");
+
+    if (vkEnumerateInstanceExtensionProperties(NULL, &ep_count, NULL) != VK_SUCCESS)
+        return;
+
+    ep = calloc(ep_count, sizeof(VkExtensionProperties));
+
+    if (vkEnumerateInstanceExtensionProperties(NULL, &ep_count, ep) != VK_SUCCESS)
+    {
+        free(ep);
+        return;
+    }
+
+    for (i = 0;  i < ep_count;  i++)
+        printf(" %s (v%u)\n", ep[i].extensionName, ep[i].specVersion);
+
+    free(ep);
+}
+
+static void list_vulkan_instance_layers(void)
+{
+    uint32_t i, lp_count = 0;
+    VkLayerProperties* lp;
+    PFN_vkEnumerateInstanceLayerProperties vkEnumerateInstanceLayerProperties =
+        (PFN_vkEnumerateInstanceLayerProperties)
+        glfwGetInstanceProcAddress(NULL, "vkEnumerateInstanceLayerProperties");
+
+    printf("Vulkan instance layers:\n");
+
+    if (vkEnumerateInstanceLayerProperties(&lp_count, NULL) != VK_SUCCESS)
+        return;
+
+    lp = calloc(lp_count, sizeof(VkLayerProperties));
+
+    if (vkEnumerateInstanceLayerProperties(&lp_count, lp) != VK_SUCCESS)
+    {
+        free(lp);
+        return;
+    }
+
+    for (i = 0;  i < lp_count;  i++)
+    {
+        printf(" %s (v%u) \"%s\"\n",
+               lp[i].layerName,
+               lp[i].specVersion >> 22,
+               lp[i].description);
+    }
+
+    free(lp);
+}
+
+static void list_vulkan_device_extensions(VkInstance instance, VkPhysicalDevice device)
+{
+    uint32_t i, ep_count;
+    VkExtensionProperties* ep;
+    PFN_vkEnumerateDeviceExtensionProperties vkEnumerateDeviceExtensionProperties =
+        (PFN_vkEnumerateDeviceExtensionProperties)
+        glfwGetInstanceProcAddress(instance, "vkEnumerateDeviceExtensionProperties");
+
+    printf("Vulkan device extensions:\n");
+
+    if (vkEnumerateDeviceExtensionProperties(device, NULL, &ep_count, NULL) != VK_SUCCESS)
+        return;
+
+    ep = calloc(ep_count, sizeof(VkExtensionProperties));
+
+    if (vkEnumerateDeviceExtensionProperties(device, NULL, &ep_count, ep) != VK_SUCCESS)
+    {
+        free(ep);
+        return;
+    }
+
+    for (i = 0;  i < ep_count;  i++)
+        printf(" %s (v%u)\n", ep[i].extensionName, ep[i].specVersion);
+
+    free(ep);
+}
+
+static void list_vulkan_device_layers(VkInstance instance, VkPhysicalDevice device)
+{
+    uint32_t i, lp_count;
+    VkLayerProperties* lp;
+    PFN_vkEnumerateDeviceLayerProperties vkEnumerateDeviceLayerProperties =
+        (PFN_vkEnumerateDeviceLayerProperties)
+        glfwGetInstanceProcAddress(instance, "vkEnumerateDeviceLayerProperties");
+
+    printf("Vulkan device layers:\n");
+
+    if (vkEnumerateDeviceLayerProperties(device, &lp_count, NULL) != VK_SUCCESS)
+        return;
+
+    lp = calloc(lp_count, sizeof(VkLayerProperties));
+
+    if (vkEnumerateDeviceLayerProperties(device, &lp_count, lp) != VK_SUCCESS)
+    {
+        free(lp);
+        return;
+    }
+
+    for (i = 0;  i < lp_count;  i++)
+    {
+        printf(" %s (v%u) \"%s\"\n",
+               lp[i].layerName,
+               lp[i].specVersion >> 22,
+               lp[i].description);
+    }
+
+    free(lp);
+}
+
+static int valid_version(void)
 {
     int major, minor, revision;
     glfwGetVersion(&major, &minor, &revision);
@@ -198,13 +333,13 @@ static GLboolean valid_version(void)
     if (major != GLFW_VERSION_MAJOR)
     {
         printf("*** ERROR: GLFW major version mismatch! ***\n");
-        return GL_FALSE;
+        return GLFW_FALSE;
     }
 
     if (minor != GLFW_VERSION_MINOR || revision != GLFW_VERSION_REVISION)
         printf("*** WARNING: GLFW version mismatch! ***\n");
 
-    return GL_TRUE;
+    return GLFW_TRUE;
 }
 
 static void print_version(void)
@@ -222,24 +357,27 @@ static void print_version(void)
 
 int main(int argc, char** argv)
 {
-    int ch, api, major, minor, revision, profile;
+    int ch, client, major, minor, revision, profile;
     GLint redbits, greenbits, bluebits, alphabits, depthbits, stencilbits;
-    GLboolean list = GL_FALSE;
+    int list_extensions = GLFW_FALSE, list_layers = GLFW_FALSE;
+    GLenum error;
     GLFWwindow* window;
 
-    enum { API, BEHAVIOR, DEBUG, FORWARD, HELP, EXTENSIONS,
+    enum { CLIENT, CONTEXT, BEHAVIOR, DEBUG, FORWARD, HELP, EXTENSIONS, LAYERS,
            MAJOR, MINOR, PROFILE, ROBUSTNESS, VERSION,
            REDBITS, GREENBITS, BLUEBITS, ALPHABITS, DEPTHBITS, STENCILBITS,
            ACCUMREDBITS, ACCUMGREENBITS, ACCUMBLUEBITS, ACCUMALPHABITS,
-           AUXBUFFERS, SAMPLES, STEREO, SRGB, SINGLEBUFFER };
+           AUXBUFFERS, SAMPLES, STEREO, SRGB, SINGLEBUFFER, NOERROR_SRSLY };
     const struct option options[] =
     {
         { "behavior",         1, NULL, BEHAVIOR },
-        { "client-api",       1, NULL, API },
+        { "client-api",       1, NULL, CLIENT },
+        { "context-api",      1, NULL, CONTEXT },
         { "debug",            0, NULL, DEBUG },
         { "forward",          0, NULL, FORWARD },
         { "help",             0, NULL, HELP },
         { "list-extensions",  0, NULL, EXTENSIONS },
+        { "list-layers",      0, NULL, LAYERS },
         { "major",            1, NULL, MAJOR },
         { "minor",            1, NULL, MINOR },
         { "profile",          1, NULL, PROFILE },
@@ -260,6 +398,7 @@ int main(int argc, char** argv)
         { "stereo",           0, NULL, STEREO },
         { "srgb",             0, NULL, SRGB },
         { "singlebuffer",     0, NULL, SINGLEBUFFER },
+        { "no-error",         0, NULL, NOERROR_SRSLY },
         { NULL, 0, NULL, 0 }
     };
 
@@ -273,12 +412,12 @@ int main(int argc, char** argv)
     if (!glfwInit())
         exit(EXIT_FAILURE);
 
-    while ((ch = getopt_long(argc, argv, "a:b:dfhlm:n:p:s:v", options, NULL)) != -1)
+    while ((ch = getopt_long(argc, argv, "a:b:c:dfhlm:n:p:s:v", options, NULL)) != -1)
     {
         switch (ch)
         {
             case 'a':
-            case API:
+            case CLIENT:
                 if (strcasecmp(optarg, API_NAME_OPENGL) == 0)
                     glfwWindowHint(GLFW_CLIENT_API, GLFW_OPENGL_API);
                 else if (strcasecmp(optarg, API_NAME_OPENGL_ES) == 0)
@@ -307,13 +446,25 @@ int main(int argc, char** argv)
                     exit(EXIT_FAILURE);
                 }
                 break;
+            case 'c':
+            case CONTEXT:
+                if (strcasecmp(optarg, API_NAME_NATIVE) == 0)
+                    glfwWindowHint(GLFW_CONTEXT_CREATION_API, GLFW_NATIVE_CONTEXT_API);
+                else if (strcasecmp(optarg, API_NAME_EGL) == 0)
+                    glfwWindowHint(GLFW_CONTEXT_CREATION_API, GLFW_EGL_CONTEXT_API);
+                else
+                {
+                    usage();
+                    exit(EXIT_FAILURE);
+                }
+                break;
             case 'd':
             case DEBUG:
-                glfwWindowHint(GLFW_OPENGL_DEBUG_CONTEXT, GL_TRUE);
+                glfwWindowHint(GLFW_OPENGL_DEBUG_CONTEXT, GLFW_TRUE);
                 break;
             case 'f':
             case FORWARD:
-                glfwWindowHint(GLFW_OPENGL_FORWARD_COMPAT, GL_TRUE);
+                glfwWindowHint(GLFW_OPENGL_FORWARD_COMPAT, GLFW_TRUE);
                 break;
             case 'h':
             case HELP:
@@ -321,7 +472,10 @@ int main(int argc, char** argv)
                 exit(EXIT_SUCCESS);
             case 'l':
             case EXTENSIONS:
-                list = GL_TRUE;
+                list_extensions = GLFW_TRUE;
+                break;
+            case LAYERS:
+                list_layers = GLFW_TRUE;
                 break;
             case 'm':
             case MAJOR:
@@ -444,13 +598,16 @@ int main(int argc, char** argv)
                     glfwWindowHint(GLFW_SAMPLES, atoi(optarg));
                 break;
             case STEREO:
-                glfwWindowHint(GLFW_STEREO, GL_TRUE);
+                glfwWindowHint(GLFW_STEREO, GLFW_TRUE);
                 break;
             case SRGB:
-                glfwWindowHint(GLFW_SRGB_CAPABLE, GL_TRUE);
+                glfwWindowHint(GLFW_SRGB_CAPABLE, GLFW_TRUE);
                 break;
             case SINGLEBUFFER:
-                glfwWindowHint(GLFW_DOUBLEBUFFER, GL_FALSE);
+                glfwWindowHint(GLFW_DOUBLEBUFFER, GLFW_FALSE);
+                break;
+            case NOERROR_SRSLY:
+                glfwWindowHint(GLFW_CONTEXT_NO_ERROR, GLFW_TRUE);
                 break;
             default:
                 usage();
@@ -460,7 +617,7 @@ int main(int argc, char** argv)
 
     print_version();
 
-    glfwWindowHint(GLFW_VISIBLE, GL_FALSE);
+    glfwWindowHint(GLFW_VISIBLE, GLFW_FALSE);
 
     window = glfwCreateWindow(200, 200, "Version", NULL, NULL);
     if (!window)
@@ -470,43 +627,50 @@ int main(int argc, char** argv)
     }
 
     glfwMakeContextCurrent(window);
+    gladLoadGLLoader((GLADloadproc) glfwGetProcAddress);
+
+    error = glGetError();
+    if (error != GL_NO_ERROR)
+        printf("*** OpenGL error after make current: 0x%08x ***\n", error);
 
     // Report client API version
 
-    api = glfwGetWindowAttrib(window, GLFW_CLIENT_API);
+    client = glfwGetWindowAttrib(window, GLFW_CLIENT_API);
     major = glfwGetWindowAttrib(window, GLFW_CONTEXT_VERSION_MAJOR);
     minor = glfwGetWindowAttrib(window, GLFW_CONTEXT_VERSION_MINOR);
     revision = glfwGetWindowAttrib(window, GLFW_CONTEXT_REVISION);
     profile = glfwGetWindowAttrib(window, GLFW_OPENGL_PROFILE);
 
     printf("%s context version string: \"%s\"\n",
-           get_api_name(api),
+           get_api_name(client),
            glGetString(GL_VERSION));
 
     printf("%s context version parsed by GLFW: %u.%u.%u\n",
-           get_api_name(api),
+           get_api_name(client),
            major, minor, revision);
 
     // Report client API context properties
 
-    if (api == GLFW_OPENGL_API)
+    if (client == GLFW_OPENGL_API)
     {
         if (major >= 3)
         {
             GLint flags;
 
             glGetIntegerv(GL_CONTEXT_FLAGS, &flags);
-            printf("%s context flags (0x%08x):", get_api_name(api), flags);
+            printf("%s context flags (0x%08x):", get_api_name(client), flags);
 
             if (flags & GL_CONTEXT_FLAG_FORWARD_COMPATIBLE_BIT)
                 printf(" forward-compatible");
-            if (flags & GL_CONTEXT_FLAG_DEBUG_BIT)
+            if (flags & 2/*GL_CONTEXT_FLAG_DEBUG_BIT*/)
                 printf(" debug");
             if (flags & GL_CONTEXT_FLAG_ROBUST_ACCESS_BIT_ARB)
                 printf(" robustness");
+            if (flags & 8/*GL_CONTEXT_FLAG_NO_ERROR_BIT_KHR*/)
+                printf(" no-error");
             putchar('\n');
 
-            printf("%s context flags parsed by GLFW:", get_api_name(api));
+            printf("%s context flags parsed by GLFW:", get_api_name(client));
 
             if (glfwGetWindowAttrib(window, GLFW_OPENGL_FORWARD_COMPAT))
                 printf(" forward-compatible");
@@ -514,6 +678,8 @@ int main(int argc, char** argv)
                 printf(" debug");
             if (glfwGetWindowAttrib(window, GLFW_CONTEXT_ROBUSTNESS) == GLFW_LOSE_CONTEXT_ON_RESET)
                 printf(" robustness");
+            if (glfwGetWindowAttrib(window, GLFW_CONTEXT_NO_ERROR))
+                printf(" no-error");
             putchar('\n');
         }
 
@@ -523,12 +689,12 @@ int main(int argc, char** argv)
             glGetIntegerv(GL_CONTEXT_PROFILE_MASK, &mask);
 
             printf("%s profile mask (0x%08x): %s\n",
-                   get_api_name(api),
+                   get_api_name(client),
                    mask,
                    get_profile_name_gl(mask));
 
             printf("%s profile mask parsed by GLFW: %s\n",
-                   get_api_name(api),
+                   get_api_name(client),
                    get_profile_name_glfw(profile));
         }
 
@@ -539,44 +705,34 @@ int main(int argc, char** argv)
             glGetIntegerv(GL_RESET_NOTIFICATION_STRATEGY_ARB, &strategy);
 
             printf("%s robustness strategy (0x%08x): %s\n",
-                   get_api_name(api),
+                   get_api_name(client),
                    strategy,
                    get_strategy_name_gl(strategy));
 
             printf("%s robustness strategy parsed by GLFW: %s\n",
-                   get_api_name(api),
+                   get_api_name(client),
                    get_strategy_name_glfw(robustness));
         }
     }
 
     printf("%s context renderer string: \"%s\"\n",
-           get_api_name(api),
+           get_api_name(client),
            glGetString(GL_RENDERER));
     printf("%s context vendor string: \"%s\"\n",
-           get_api_name(api),
+           get_api_name(client),
            glGetString(GL_VENDOR));
 
     if (major >= 2)
     {
         printf("%s context shading language version: \"%s\"\n",
-               get_api_name(api),
+               get_api_name(client),
                glGetString(GL_SHADING_LANGUAGE_VERSION));
     }
 
-    printf("Framebuffer:\n");
+    printf("%s framebuffer:\n", get_api_name(client));
 
-    if (api == GLFW_OPENGL_API && profile == GLFW_OPENGL_CORE_PROFILE)
+    if (client == GLFW_OPENGL_API && profile == GLFW_OPENGL_CORE_PROFILE)
     {
-        PFNGLGETFRAMEBUFFERATTACHMENTPARAMETERIVPROC
-            glGetFramebufferAttachmentParameteriv =
-            (PFNGLGETFRAMEBUFFERATTACHMENTPARAMETERIVPROC)
-            glfwGetProcAddress("glGetFramebufferAttachmentParameteriv");
-        if (!glGetFramebufferAttachmentParameteriv)
-        {
-            glfwTerminate();
-            exit(EXIT_FAILURE);
-        }
-
         glGetFramebufferAttachmentParameteriv(GL_FRAMEBUFFER,
                                               GL_BACK_LEFT,
                                               GL_FRAMEBUFFER_ATTACHMENT_RED_SIZE,
@@ -615,7 +771,7 @@ int main(int argc, char** argv)
     printf(" red: %u green: %u blue: %u alpha: %u depth: %u stencil: %u\n",
            redbits, greenbits, bluebits, alphabits, depthbits, stencilbits);
 
-    if (api == GLFW_OPENGL_ES_API ||
+    if (client == GLFW_OPENGL_ES_API ||
         glfwExtensionSupported("GL_ARB_multisample") ||
         major > 1 || minor >= 3)
     {
@@ -626,7 +782,7 @@ int main(int argc, char** argv)
         printf(" samples: %u sample buffers: %u\n", samples, samplebuffers);
     }
 
-    if (api == GLFW_OPENGL_API && profile != GLFW_OPENGL_CORE_PROFILE)
+    if (client == GLFW_OPENGL_API && profile != GLFW_OPENGL_CORE_PROFILE)
     {
         GLint accumredbits, accumgreenbits, accumbluebits, accumalphabits;
         GLint auxbuffers;
@@ -641,9 +797,101 @@ int main(int argc, char** argv)
                accumredbits, accumgreenbits, accumbluebits, accumalphabits, auxbuffers);
     }
 
-    // Report client API extensions
-    if (list)
-        list_extensions(api, major, minor);
+    if (list_extensions)
+        list_context_extensions(client, major, minor);
+
+    printf("Vulkan loader: %s\n",
+           glfwVulkanSupported() ? "available" : "missing");
+
+    if (glfwVulkanSupported())
+    {
+        uint32_t i, re_count, pd_count;
+        const char** re;
+        VkApplicationInfo ai = {0};
+        VkInstanceCreateInfo ici = {0};
+        VkInstance instance;
+        VkPhysicalDevice* pd;
+        PFN_vkCreateInstance vkCreateInstance = (PFN_vkCreateInstance)
+            glfwGetInstanceProcAddress(NULL, "vkCreateInstance");
+        PFN_vkDestroyInstance vkDestroyInstance;
+        PFN_vkEnumeratePhysicalDevices vkEnumeratePhysicalDevices;
+        PFN_vkGetPhysicalDeviceProperties vkGetPhysicalDeviceProperties;
+
+        re = glfwGetRequiredInstanceExtensions(&re_count);
+
+        printf("Vulkan required instance extensions:");
+        for (i = 0;  i < re_count;  i++)
+            printf(" %s", re[i]);
+        putchar('\n');
+
+        if (list_extensions)
+            list_vulkan_instance_extensions();
+
+        if (list_layers)
+            list_vulkan_instance_layers();
+
+        ai.sType = VK_STRUCTURE_TYPE_APPLICATION_INFO;
+        ai.pApplicationName = "glfwinfo";
+        ai.applicationVersion = GLFW_VERSION_MAJOR;
+        ai.pEngineName = "GLFW";
+        ai.engineVersion = GLFW_VERSION_MAJOR;
+        ai.apiVersion = VK_API_VERSION_1_0;
+
+        ici.sType = VK_STRUCTURE_TYPE_INSTANCE_CREATE_INFO;
+        ici.pApplicationInfo = &ai;
+        ici.enabledExtensionCount = re_count;
+        ici.ppEnabledExtensionNames = re;
+
+        if (vkCreateInstance(&ici, NULL, &instance) != VK_SUCCESS)
+        {
+            glfwTerminate();
+            exit(EXIT_FAILURE);
+        }
+
+        vkDestroyInstance = (PFN_vkDestroyInstance)
+            glfwGetInstanceProcAddress(instance, "vkDestroyInstance");
+        vkEnumeratePhysicalDevices = (PFN_vkEnumeratePhysicalDevices)
+            glfwGetInstanceProcAddress(instance, "vkEnumeratePhysicalDevices");
+        vkGetPhysicalDeviceProperties = (PFN_vkGetPhysicalDeviceProperties)
+            glfwGetInstanceProcAddress(instance, "vkGetPhysicalDeviceProperties");
+
+        if (vkEnumeratePhysicalDevices(instance, &pd_count, NULL) != VK_SUCCESS)
+        {
+            vkDestroyInstance(instance, NULL);
+            glfwTerminate();
+            exit(EXIT_FAILURE);
+        }
+
+        pd = calloc(pd_count, sizeof(VkPhysicalDevice));
+
+        if (vkEnumeratePhysicalDevices(instance, &pd_count, pd) != VK_SUCCESS)
+        {
+            free(pd);
+            vkDestroyInstance(instance, NULL);
+            glfwTerminate();
+            exit(EXIT_FAILURE);
+        }
+
+        for (i = 0;  i < pd_count;  i++)
+        {
+            VkPhysicalDeviceProperties pdp;
+
+            vkGetPhysicalDeviceProperties(pd[i], &pdp);
+
+            printf("Vulkan %s device: \"%s\"\n",
+                   get_device_type_name(pdp.deviceType),
+                   pdp.deviceName);
+
+            if (list_extensions)
+                list_vulkan_device_extensions(instance, pd[i]);
+
+            if (list_layers)
+                list_vulkan_device_layers(instance, pd[i]);
+        }
+
+        free(pd);
+        vkDestroyInstance(instance, NULL);
+    }
 
     glfwTerminate();
     exit(EXIT_SUCCESS);
